import 'dart:async';
import 'dart:convert';

//美颜使用
// import 'package:agora_rtc/rtc_engine.dart';
// import 'package:agora_rtc/rtc_local_view.dart' as rtc_local_view;
// import 'package:agora_rtc/rtc_remote_view.dart' as rtc_remote_view;
//普通使用
import 'package:agora_rtc_engine/rtc_engine.dart';
import 'package:agora_rtc_engine/rtc_local_view.dart' as rtc_local_view;
import 'package:agora_rtc_engine/rtc_remote_view.dart' as rtc_remote_view;
import 'package:agora_rtm/agora_rtm.dart';
import 'package:flutter/material.dart';
import 'package:flutter_screenutil/flutter_screenutil.dart';
import 'package:mmoo_base/api/address.dart';
import 'package:mmoo_base/config/app_config.dart';
import 'package:mmoo_base/entity/match_entity.dart';
import 'package:mmoo_base_video/entity/call_entity.dart';
import 'package:mmoo_base_video/entity/video_message.dart';

///通话逻辑-2个主视图
class VideoCallWidget extends StatefulWidget {
  final String agoraAppKey; //声网key
  final AgoraRtmClient rtmClient; //本地Rtm对象
  final String userId; //对方id
  final String name; //自己姓名
  final String portraitUri; //自己头像
  final int meUid; //自己uid
  final bool call; //主动拨打还是被动
  final bool direct; //主动拨打下直接播放，还是等待数据返回拨号
  final int type; //1:匹配来电,2:玩具来电
  final int? toyCost; //2玩具价格
  final MatchEntity? matchEntity; //匹配对象
  final int videoMatchTime; //匹配等待时间
  final int videoMatchDownTime; //匹配成功挂断提示最后倒计时间
  final double bottomBarHeight; //底部安全距离
  final bool gui; //是否正式模式
  final VideoCallDelegate? delegate; //回调方法
  final Widget? closeWidget; //关闭自己视频的占位
  final bool position; //默认占位
  final double? left;
  final double? top;
  final double? right;
  final double? bottom;
  final double? borderRadius;

  const VideoCallWidget(Key key, this.agoraAppKey, this.rtmClient, this.userId,
      this.name, this.portraitUri, this.meUid, this.call, this.direct,
      {this.delegate,
      this.closeWidget,
      this.type = 0,
      this.toyCost,
      this.matchEntity,
      this.videoMatchTime = 0,
      this.videoMatchDownTime = 0,
      this.bottomBarHeight = 0,
      this.gui = false,
      this.position = false,
      this.left,
      this.top,
      this.right,
      this.bottom,
      this.borderRadius})
      : super(key: key);

  @override
  VideoCallWidgetState createState() => VideoCallWidgetState();
}

class VideoCallWidgetState extends State<VideoCallWidget> {
  RtcEngine? rtcEngine;
  int? userUid; //对方uid
  int type = 0; //1:匹配来电,2:玩具来电
  bool call = false; //是否主叫
  bool match15 = false; //是否接通进行15s
  bool matchSoon = false; //ppei中通话不足1分钟
  bool isSoonMoney = false; //快没钱了
  bool? isConnect; //是否链接中，处理callStatus==0的时候文字显示问题
  bool isSendConnect = false; //是否发送邀请
  bool offline = false; //对方是否已经挂电话了
  //-1-未点击拨号(获取到数据显示待接听或拨号界面) 0-链接中 1-正在通话 2-已挂断 3-没钱已挂断 4-通话前置没钱-免费通话
  int callStatus = -1;
  bool localVideoStatus = false; //是否本地推流成功
  bool remoteVideoStatus = false; //是否远端推流成功
  int currentRate = 0; //计时
  bool showTime = false; //显示计时器
  bool partyLeft = false; //对方是否挂断
  int currentDown = 0; //通话结束计时
  //网络质量检测
  int quality = 0;
  int qualityDown = 0;
  bool showVideoView = true; //显示界面其他按钮
  ///用户什么时候加入房间
  int userJoinedRate = 0;
  bool isLeave = false; //是否离开房间
  String? rtcToken; //房间号
  String? channelId; //房间号
  String? callId; //通话id

  //ui控制
  bool closeCamera = false; //关闭摄像头
  bool closeMicrophone = false; //关闭麦克风
  double bottomHeight = 0; //键盘或者其他界面弹起时界面上移的距离

  @override
  void initState() {
    super.initState();
    type = widget.type;
    call = widget.call;
    addRtmHandler();
    initRtcEngine();
  }

  @override
  void dispose() {
    rtcEngine?.stopLastmileProbeTest();
    rtcEngine?.leaveChannel();
    super.dispose();
  }

  void setBottomHeight(double bottomHeight) {
    this.bottomHeight = bottomHeight;
    setState(() {});
  }

  bool setShowVideoView() {
    showVideoView = !showVideoView;
    setState(() {});
    return showVideoView;
  }

  Future<bool> setCloseCamera() async {
    closeCamera = !closeCamera;
    if (callStatus == 1) {
      await rtcEngine?.muteLocalVideoStream(closeCamera);
    }
    if (!mounted) {
      return false;
    }
    setState(() {});
    return closeCamera;
  }

  Future<bool> setCloseMicrophone() async {
    closeMicrophone = !closeMicrophone;
    if (callStatus == 1) {
      await rtcEngine?.enableLocalAudio(!closeMicrophone);
    }
    return closeMicrophone;
  }

  void switchCamera() {
    if (callStatus == 1) {
      rtcEngine?.switchCamera();
    }
  }

  void restCallStatus() {
    if (type == 1) {
      type = 0;
    }
    match15 = false;
    matchSoon = false;
    call = true;
    callStatus = -1;
    isSoonMoney = false;
    isConnect = null;
    isSendConnect = false;
    offline = false;
    localVideoStatus = false;
    remoteVideoStatus = false;
    showTime = false;
    partyLeft = false;
    showVideoView = true;
    isLeave = false;
    closeCamera = false;
    closeMicrophone = false;
    userUid = null;
    channelId = null;
    setState(() {});
  }

  void setCallStatus(int callStatus) {
    this.callStatus = callStatus;
    setState(() {});
  }

  void onSetState() {
    setState(() {});
  }

  @override
  Widget build(BuildContext context) {
    return callStatus <= 2 && userUid != null && channelId != null
        ? Stack(
            children: [
              rtc_remote_view.SurfaceView(uid: userUid!, channelId: channelId!),
              Visibility(
                visible: showVideoView,
                child: Positioned(
                  left: widget.left,
                  top: widget.top,
                  right: widget.right ??
                      (widget.position ? ScreenUtil().setWidth(15) : null),
                  bottom: widget.bottom ??
                      (widget.position
                          ? ScreenUtil().setWidth(80) +
                              widget.bottomBarHeight +
                              bottomHeight
                          : null),
                  child: SizedBox(
                    width: ScreenUtil().setWidth(100),
                    height: ScreenUtil().setWidth(150),
                    child: ClipRRect(
                      borderRadius:
                          BorderRadius.circular(widget.borderRadius ?? 8),
                      child: Stack(
                        fit: StackFit.expand,
                        children: <Widget>[
                          rtc_local_view.SurfaceView(),
                          Visibility(
                              visible:
                                  widget.closeWidget != null && closeCamera,
                              child: widget.closeWidget ?? Container()),
                        ],
                      ),
                    ),
                  ),
                ),
              ),
            ],
          )
        : Container();
  }

  void addRtmHandler() async {
    widget.rtmClient.onLocalInvitationReceivedByPeer =
        (AgoraRtmLocalInvitation invite) {
      currentRate = 0;
      widget.delegate?.onVideoSetState();
    };
    widget.rtmClient.onLocalInvitationAccepted =
        (AgoraRtmLocalInvitation invite) {
      joinChannel();
    };
    widget.rtmClient.onLocalInvitationRefused =
        (AgoraRtmLocalInvitation invite) {
      if (call) {
        partyLeft = true;
        widget.delegate?.onVideoSetState();
      } else {
        widget.delegate?.onShowToast(5);
      }
      widget.delegate?.onNormalPop(delayed: call);
    };
    widget.rtmClient.onRemoteInvitationCanceled =
        (AgoraRtmRemoteInvitation invite) {
      widget.delegate?.onShowToast(5);
      widget.delegate?.onNormalPop();
    };
  }

  void userInHouseCheck() async {
    try {
      await widget.rtmClient
          .acceptRemoteInvitation({'callerId': widget.userId});
    } catch (e) {
      printLog(e);
    }
    joinChannel();
  }

  void joinChannel() async {
    try {
      await rtcEngine?.stopLastmileProbeTest();
      await rtcEngine?.joinChannel(rtcToken, channelId!, null, widget.meUid);
      isConnect = true;
      callStatus = 0;
      currentRate = 0;
      widget.delegate?.onCallStatus(callStatus);
      if (!mounted) {
        return;
      }
      setState(() {});
    } catch (e) {
      printLog(e);
      widget.delegate?.onShowToast(-1);
      willSendVideo(1);
      widget.delegate?.onNormalPop();
    }
  }

  void addAgoraEventHandlers() {
    RtcEngineEventHandler handler = RtcEngineEventHandler();
    handler.tokenPrivilegeWillExpire = (String token) async {
      printLog('tokenPrivilegeWillExpire: $token');
      String? rtcToken = await widget.delegate?.getRtcToken(channelId!);
      if (rtcToken?.isEmpty == false) {
        rtcEngine?.renewToken(rtcToken!);
      }
    };
    handler.lastmileQuality = (NetworkQuality quality) {
      printLog('lastmileQuality: ${quality.index}');
      if (quality == NetworkQuality.Unknown ||
          quality == NetworkQuality.Excellent ||
          quality == NetworkQuality.Good ||
          quality == NetworkQuality.Poor ||
          quality == NetworkQuality.Bad) {
        if (call && callStatus != 0) {
          callStatus = 0;
          willSendVideo(0);
        } else {
          callStatus = -1;
        }
      } else {
        widget.delegate?.onShowToast(0);
        willSendVideo(1);
        widget.delegate?.onNormalPop();
      }
    };
    handler.joinChannelSuccess = (String channel, int uid, int elapsed) {
      printLog('onJoinChannelSuccess: $channel, uid: $uid, elapsed: $elapsed');
    };
    handler.userJoined = (int uid, int elapsed) {
      printLog('onUserJoined: uid: $uid, elapsed: $elapsed');
      userJoinedRate = DateTime.now().millisecondsSinceEpoch;
    };
    handler.localVideoStateChanged =
        (LocalVideoStreamState state, LocalVideoStreamError error) async {
      printLog('onLocalVideoStateChanged: $state $error');
      if (state == LocalVideoStreamState.Encoding) {
        localVideoStatus = true;
      } else {
        widget.delegate?.onShowLocalVideo(true);
      }
    };
    handler.firstRemoteVideoFrame =
        (int uid, int width, int height, int elapsed) {
      printLog('firstRemoteVideoFrame: $uid $elapsed');
      if (callStatus == 0) {
        callStatus = 1;
        widget.delegate?.onCallStatus(callStatus);
        if (type == 1) {
          matchSuccess();
        } else {
          callStart();
        }
      }
    };
    handler.remoteVideoStateChanged = (int uid, VideoRemoteState state,
        VideoRemoteStateReason reason, int elapsed) {
      printLog('onRemoteVideoStateChanged: $uid $state $reason $elapsed');
      if (state == VideoRemoteState.Decoding) {
        remoteVideoStatus = true;
        if (userUid == null) {
          userUid = uid;
          if (!mounted) {
            return;
          }
          setState(() {});
        }
      } else {
        widget.delegate?.onShowLocalVideo(false);
      }
    };
    handler.networkQuality =
        (int uid, NetworkQuality txQuality, NetworkQuality rxQuality) {
      // printLog('onNetworkQuality: $uid $txQuality $rxQuality');
      if (callStatus != 1) {
        return;
      }
      if (txQuality == NetworkQuality.Down ||
          rxQuality == NetworkQuality.Down) {
        qualityDown++;
        if (qualityDown >= 3) {
          qualityDown = 0;
          widget.delegate?.onShowToast(0);
          if (!offline) {
            widget.delegate?.onRemoteUserLeft(0, toast: false);
          }
        }
      }
      //下行网络
      if (rxQuality == NetworkQuality.Excellent ||
          rxQuality == NetworkQuality.Good ||
          rxQuality == NetworkQuality.Poor) {
        quality++;
        if (quality >= 3) {
          quality = 0;
          widget.delegate?.onShowNetText(null);
        }
      } else if (rxQuality == NetworkQuality.Unknown ||
          rxQuality == NetworkQuality.Bad ||
          rxQuality == NetworkQuality.VBad) {
        quality = 0;
        widget.delegate?.onShowNetText(uid == 0);
      }
    };
    handler.userOffline = (int uid, UserOfflineReason reason) {
      printLog('onUserOffline: $uid, reason: $reason');
      if (isSoonMoney && currentRate % 60 >= 55) {
        return;
      }
      if (offline) {
        return;
      }
      offline = true;
      widget.delegate
          ?.onRemoteUserLeft(reason == UserOfflineReason.Dropped ? 3 : 1);
    };
    handler.leaveChannel = (RtcStats stats) {
      printLog('onLeaveChannel');
      if (callStatus != 1) {
        return;
      }
      if (!offline) {
        widget.delegate?.onRemoteUserLeft(0);
      }
    };
    handler.error = (ErrorCode err) {
      printLog('onError: $err, index: ${err.index}');
      if (err.index < 1000 || err.index > 1400) {
        if (callStatus == 0) {
          if (err == ErrorCode.ClientIsBannedByServer) {
            widget.delegate?.onShowToast(1);
          } else {
            widget.delegate?.onShowToast(2);
          }
          willSendVideo(1);
          widget.delegate?.onNormalPop();
        } else if (callStatus == 1) {
          widget.delegate?.onRemoteUserLeft(4);
        } else {
          widget.delegate?.onNormalPop(hangup: false);
        }
      }
    };
    handler.rtcStats = (RtcStats stats) {
      if (!mounted) {
        return;
      }
      // printLog(
      //     'onRtcStats:${stats.totalDuration}, lastmileDelay:${stats.lastmileDelay}');
      if (showTime && callStatus == 1) {
        currentCall();
        Future.delayed(const Duration(seconds: 1), () {
          if (!mounted) {
            return;
          }
          currentCall();
        });
      }
    };
    rtcEngine?.setEventHandler(handler);
  }

  void initRtcEngine() async {
    //美颜使用
    // rtcEngine = await RtcEngine.create(widget.agoraAppKey);
    //普通使用
    rtcEngine =
        await RtcEngine.createWithContext(RtcEngineContext(widget.agoraAppKey));
    addAgoraEventHandlers();
    rtcEngine?.enableVideo();
    rtcEngine?.enableAudio();
    rtcEngine?.setChannelProfile(ChannelProfile.Communication);
    VideoEncoderConfiguration config = VideoEncoderConfiguration();
    config.orientationMode = VideoOutputOrientationMode.FixedPortrait;
    rtcEngine?.setVideoEncoderConfiguration(config);
    widget.delegate?.onVideoSetState();
  }

  void callBefore() async {
    LastmileProbeConfig probeConfig =
        LastmileProbeConfig(true, true, 100000, 100000);
    rtcEngine?.startLastmileProbeTest(probeConfig);
  }

  void currentCall() {
    if (!mounted) {
      return;
    }
    currentRate++;
    if (currentRate == 10 && !localVideoStatus) {
      widget.delegate?.onShowLocalVideo(true, check: false);
    }
    if (currentRate == 20 && !remoteVideoStatus) {
      widget.delegate?.onShowLocalVideo(false, check: false);
    }
    if (matchSoon && match15 && currentRate == widget.videoMatchTime ~/ 3) {
      widget.delegate?.onShowMatchWidget();
    }
    if (match15 && currentRate == widget.videoMatchTime) {
      if (matchSoon) {
        widget.delegate?.onShowToast(3);
        widget.delegate?.showEndView(true);
      } else {
        matchStart();
      }
    }
    if (match15 &&
        !matchSoon &&
        !showTime &&
        currentRate == widget.videoMatchTime + 5) {
      widget.delegate?.onShowToast(3);
      widget.delegate?.showEndView(true);
    }
    if (currentRate % 60 == 0) {
      if (isSoonMoney) {
        isSoonMoney = false;
        widget.delegate?.onCalling(currentRate - 1);
      } else {
        widget.delegate?.onCalling(currentRate);
      }
    }
    if (currentDown > 0) {
      currentDown--;
      widget.delegate?.onCurrentRate(currentDown, down: true);
      if (currentDown == widget.videoMatchDownTime) {
        widget.delegate?.onVideoSetState();
      }
      if (!match15 &&
          currentDown <= widget.videoMatchDownTime &&
          currentRate % 2 == 0) {
        widget.delegate?.onCurrentRate(currentRate, show: true);
      }
    }
    widget.delegate?.onCurrentRate(currentRate);
  }

  void matchSuccess() async {
    await widget.delegate?.onCallStart(type);
    currentDown = widget.videoMatchTime;
    match15 = true;
    willSendVideo(5,
        newTime: DateTime.now().millisecondsSinceEpoch - userJoinedRate);
    currentRate = 0;
    showTime = true;
    if (!mounted) {
      return;
    }
    setState(() {});
  }

  void matchStart() async {
    willSendVideo(9);
    match15 = false;
    currentDown = 0;
    await callStart();
    widget.delegate?.onShowToast(4);
  }

  Future callStart() async {
    if (type == 1) {
      type = 0;
    }
    await widget.delegate?.onCallStart(type);
    willSendVideo(5,
        newTime: DateTime.now().millisecondsSinceEpoch - userJoinedRate);
    currentRate = 0;
    showTime = true;
    if (!mounted) {
      return;
    }
    setState(() {});
    widget.delegate?.onCalling(currentRate);
  }

  void showMoney(bool rest) {
    if (!widget.gui) {
      return;
    }
    int? size = widget.delegate?.onShowMoney();
    isSoonMoney = size == 0;
    // if (size <= 1) {
    if (size == 0) {
      if (currentDown == 0 || rest) {
        currentDown = (size! + 1) * 60 - currentRate % 60;
      }
    } else {
      currentDown = 0;
    }
    if (size == 1) {
      printLog('通话中费用不足2分钟');
      willSendVideo(8);
      // if (currentRate >= 60 || currentRate < 60 && !call) {
      //   showRechargeDialog(call: false, rest: false, min: '2');
      // }
    } else if (size == 0) {
      printLog('通话中费用不足1分钟');
      willSendVideo(6);
      if (currentRate > 60) {
        widget.delegate?.onShowMatchWidget(recharge: true);
      }
    }
  }

  void willSendVideo(int status, {int? newTime}) async {
    try {
      printLog('willSendVideo: $status');
      if (widget.delegate?.onWillCheckSendVideo() == false) {
        leaveChannel();
        return;
      }
      if (status == 0) {
        if (callStatus < 0 && call) {
          return;
        }
        var message = CallEntity();
        message.callId = callId;
        message.userName = widget.name;
        message.userImage = widget.portraitUri;
        message.callPackage = Address.packageName;
        message.callType = type;
        if (type == 1) {
          message.callMatch = true;
          message.matchCostUserId = widget.matchEntity!.matchCostUserId;
          message.callCostUserId = widget.matchEntity!.callCostUserId;
          message.callCost = widget.matchEntity!.callCost;
          message.matchTime = widget.matchEntity!.matchTime;
        } else if (type == 2) {
          message.toyCost = widget.toyCost;
        }
        await widget.rtmClient.sendLocalInvitation({
          'calleeId': widget.userId,
          'channelId': channelId,
          'content': jsonEncode(message.toJson())
        });
        isSendConnect = true;
      } else if (status == 1 && (isConnect ?? false) == false) {
        if (callStatus < 0 && call) {
          return;
        }
        if (call) {
          var message = CallEntity();
          message.callId = callId;
          message.userName = widget.name;
          message.userImage = widget.portraitUri;
          message.callPackage = Address.packageName;
          await widget.rtmClient.cancelLocalInvitation({
            'calleeId': widget.userId,
            'channelId': channelId,
            'content': jsonEncode(message.toJson())
          });
        } else {
          await widget.rtmClient
              .refuseRemoteInvitation({'callerId': widget.userId});
        }
      } else {
        if (status == 1 && callStatus == 1) {
          return;
        }
        VideoMessage videoMsg = VideoMessage();
        videoMsg.userId = '${widget.meUid}';
        videoMsg.status = status;
        videoMsg.newTime = newTime;
        AgoraRtmMessage message =
            AgoraRtmMessage.fromText(jsonEncode(videoMsg.toJson()));
        await widget.rtmClient.sendMessageToPeer(widget.userId, message, true);
      }
    } catch (e) {
      printLog(e);
    }
  }

  Future leaveChannel([bool keep = false]) async {
    if (callStatus == 1 || callStatus == 2) {
      if (!isLeave) {
        isLeave = true;
        await rtcEngine?.leaveChannel();
      }
      if (!keep && !offline) {
        widget.delegate?.onRemoteUserLeft(0);
      }
    }
  }

  void printLog(e) {
    AppConfig.printLog(e);
  }
}

abstract class VideoCallDelegate {
  //rtm通话自定义消息-发送前检测
  bool onWillCheckSendVideo();

  //本地推流或对方推流不成功提示及挂断
  void onShowLocalVideo(bool local, {bool check = true});

  //通话状态维护
  void onCallStatus(int callStatus);

  //挂断电话
  ///0-直接挂断
  void onRemoteUserLeft(int dropped, {bool toast = true});

  //toast提示
  ///-1-错误提示
  ///0-网络不好导致直接挂断-The network has been disconnected and the call is over!
  ///1-服务器封杀-Please try again after 60s!
  ///2-其他原因-Sorry, please try again later!
  ///3-匹配结束-Match end
  ///4-匹配继续-Pay per min if the call gets started!
  ///5-对方取消通话-The other party has cancelled!
  void onShowToast(int type);

  //网络不好提示为null网络状态良好
  void onShowNetText(bool? me);

  //直接关闭页面false走挂断流程
  void onNormalPop({bool hangup = true, bool? delayed = false});

  //匹配余额不足弹窗true余额不足弹窗
  void onShowMatchWidget({bool recharge = false});

  //通话计时器-down=true倒计时通话计时器-show=true匹配进入最后20s倒计时，每2s回调一次
  void onCurrentRate(int rate, {bool down = false, bool show = false});

  //通话开始-1:匹配来电-匹配开始,2:玩具来电
  Future onCallStart(int type);

  //通话打点
  void onCalling(int currentRate);

  //通话费用判断-剩余分钟数
  int onShowMoney();

  //刷新界面
  void onVideoSetState();

  //通话结束、keep保持页面
  void showEndView([bool keep = false]);

  //token刷新
  Future<String?> getRtcToken(String channel);
}
